package com.nisum.Employeeinfo.dao;

import com.nisum.Employeeinfo.errorcode.ApiErrorCode;
import com.nisum.Employeeinfo.exception.EmployeeNotFoundException;
import com.nisum.Employeeinfo.model.Employee;
import org.slf4j.Logger;

import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.r2dbc.BadSqlGrammarException;
import org.springframework.r2dbc.core.DatabaseClient;
import org.springframework.stereotype.Repository;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@Repository
public class EmployeeDao {

    private static final Logger log = LoggerFactory.getLogger(EmployeeDao.class);

    @Autowired
    private DatabaseClient databaseClient;

    public Mono<Employee> saveEmployee(Employee employee){
        String query="INSERT INTO employee ( name, email, dept, sal) VALUES (:name, :email, :dept, :sal)";
        return databaseClient.sql(query)

                .bind("name",employee.getName())
                .bind("email",employee.getEmail())
                .bind("dept",employee.getDept())
                .bind("sal",employee.getSal())
                .fetch()
                .rowsUpdated()
//                .flatMap(rows -> {
//                    if (rows > 0) {
//                        // Successfully inserted the employee
//                        return Mono.just(employee)
//                                .doOnSuccess(savedEmployee -> {
//                                    // Side effect: log successful insert
//                                    log.info("Successfully saved employee: {}", savedEmployee);
//                                });
//                    } else {
//                        // No rows updated, employee not found (error scenario)
//                        return Mono.error(new EmployeeNotFoundException(ApiErrorCode.EMPLOYEE_NOT_FOUND, "Employee Not Found"));
//                    }
//                })
//                .onErrorResume(RuntimeException.class,ex->{
//                    log.error("invalid employee data {}",ex.getMessage());
//                    return Mono.just(new Employee(-1,"invalid name","invalid email","invalid dept",0.0));
//                });
                .flatMap(rows -> Mono.justOrEmpty(rows > 0 ? employee : null))
                .switchIfEmpty(Mono.error(new RuntimeException("no record updated")))

                // Handle runtime exceptions
                .onErrorResume(RuntimeException.class, ex -> {
                    log.error("Invalid employee data: {}", ex.getMessage());
                    return Mono.just(new Employee(-1, "invalid name", "invalid email", "invalid dept", 0.0));
                })

                // Log successful insertions
                .doOnSuccess(savedEmployee -> log.info("Successfully saved employee: {}", savedEmployee));
    }



    public Flux<Employee> findAllEmployee(){
        String query="SELECT * FROM employee ";
        return databaseClient.sql(query)
                .map((row,metadata)-> new Employee(
                        row.get("id",Integer.class),
                        row.get("name", String.class),
                        row.get("email",String.class),
                        row.get("dept",String.class),
                        row.get("sal",Double.class)
                ))
                .all()
                .doOnComplete(()->log.info("Successfully fetched all the employees"))
                .doOnError(err->log.error("Error in Fetching the Employee Records {}",err.getMessage()))
                .onErrorResume(BadSqlGrammarException.class, ex->{
                    log.error("No data found {}",ex.getMessage());
                    return Flux.empty();
                });

    }

    public Mono<Employee> findEmployeeById(Integer id){
        String query= "SELECT id,name,email,dept,sal FROM employee WHERE id = :id";

        return databaseClient.sql(query)
                .bind("id",id)
                .map((row, metadata) -> new Employee(
                        row.get("id", Integer.class),
                        row.get("name", String.class),
                        row.get("email", String.class),
                        row.get("dept", String.class),
                        row.get("sal", Double.class)
                ))
                .one()
                .doOnSuccess(employee -> log.info("Successfully fetched employee: {}", employee))
                .doOnError(error -> log.error("Error fetching employee with ID {}: {}", id, error.getMessage()))
                .onErrorResume(EmptyResultDataAccessException.class, ex -> {
                    // Handle the case where no employee is found (returning empty)
                    log.warn("No employee found with ID: {}", id);
                    return Mono.empty();  // Return an empty Mono
                });
    }
    public Mono<Employee> updateEmployee(Integer id,Employee employee){
        String query = "UPDATE employee SET name = :name, email = :email, dept = :dept, sal = :sal WHERE id = :id";

        return databaseClient
                .sql(query)
                .bind("id", id)
                .bind("name", employee.getName())
                .bind("email", employee.getEmail())
                .bind("dept", employee.getDept())
                .bind("sal", employee.getSal())
                .fetch()
                .rowsUpdated()
                .flatMap(updatedRows -> {
                    if (updatedRows > 0) {
                        // Return the updated employee object if update is successful
                        return Mono.just(employee);
                    } else {
                        // Return an error if no rows were updated (e.g., employee not found)
                        return Mono.error(new RuntimeException("Employee with ID " + id + " not found"));
                    }
                })
                .doOnSuccess(updatedEmployee -> log.info("Employee updated successfully: {}", updatedEmployee))
                .doOnError(error -> log.error("Error updating employee with ID {}: {}", id, error.getMessage()))
                .onErrorReturn(RuntimeException.class, new Employee(0, "Unknown", "unknown@domain.com", "Unknown", 0.0))
                .onErrorContinue((throwable, o) -> {
                    if (throwable instanceof RuntimeException) {
                        // Log and continue if a RuntimeException occurs
                        log.warn("Continuing despite error: {}", throwable.getMessage());
                    }
                });
    }

    public Mono<Long> deleteEmployeeById(Integer id) {
        String query = "DELETE FROM employee WHERE id = :id";

        return databaseClient.sql(query)
                .bind("id", id)
                .fetch()
                .rowsUpdated()
                .flatMap(updatedRows -> {
                    if (updatedRows > 0) {
                        // Return the number of rows deleted
                        return Mono.just(updatedRows);
                    } else {
                        // Throw an error if no rows were deleted
                        return Mono.error(new EmployeeNotFoundException(
                                ApiErrorCode.EMPLOYEE_NOT_FOUND, "Employee Not Found"));
                    }
                });
    }


}
